# coding=utf-8
#
# Copyright 2021-2022 Romain BESSEAU <romain.besseau@ec.europa.eu>
# Copyright 2022-2024 Alejandra CUE GONZALEZ <alejandra.cue_gonzalez@minesparis.psl.eu>
# Copyright 2022-2024 Benoît GSCHWIND <benoit.gschwind@minesparis.psl.eu>
# Copyright 2022-2024 MINES Paris
#
# This file is part of "parasol-lca" and you can used under the
# term of European Union Public Licence.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# European Union Public Licence for more details.
#
# Version: 22 April 2025
#
# Code authors: Romain BESSEAU, Benoît GSCHWIND, Alejandra CUE GONZALEZ
#

from collections import defaultdict

def load_ecoinvent_biosphere(path, biosphere_name):
    """
    Load ecoinvent biosphere from file, and define it as biosphere database.
    Parameters:
    -----------
        path : path to the `MasterData/ElementaryExchanges.xml' provided by file such as `ecoinvent 3.9_cutoff_ecoSpold02.7z'
        biosphere_name : name of the database where data will be stored
    """
    import os
    import brightway2 as bw
    import bw2data
    from bw2io.importers import Ecospold2BiosphereImporter, SingleOutputEcospold2Importer
    if biosphere_name not in bw.databases:
        eb = Ecospold2BiosphereImporter(name=biosphere_name,
             #filepath=os.path.join(path, "MasterData", "ElementaryExchanges.xml"))
             filepath=path)
        eb.apply_strategies()
        if not eb.all_linked:
                raise ValueError(f"Can't ingest biosphere database {biosphere_name} - unlinked flows.")
        eb.write_database(overwrite=False)
    bw2data.preferences["biosphere_database"] = biosphere_name

def load_ecoinvent_lci(path : str, db_name : str, biosphere_name : str):
    """
    Load ecoinvent LCI in db_name
    Parameters:
    -----------
        path : path to the dataset directory, usually the `datasets' directory provided by file such as `ecoinvent 3.9_cutoff_ecoSpold02.7z'
        db_name: database name where LCI will be stored
        biosphere_name: the biosphere database name to use to link LCI, be carefull to use the biosphere that corespond to LCI
    """
    import os
    import brightway2 as bw
    import bw2data
    from bw2io.importers import Ecospold2BiosphereImporter, SingleOutputEcospold2Importer
    if db_name not in bw.databases:
        soup = SingleOutputEcospold2Importer(
            #dirpath=os.path.join(path, "datasets"),
            dirpath=path,
            db_name=db_name,
            biosphere_database_name=biosphere_name,
            signal=None)
        soup.apply_strategies()
        if not soup.all_linked:
            raise ValueError(f"Can't ingest inventory database {db_name} - unlinked flows.")
        soup.write_database()

def load_ecoinvent_lcia(path : str, biosphere_name : str, version : str):
    """
    Load LCIA methods from ecoinvent file
    Parameters:
    ----------
        path: path to the ecoinvent xlsx file, such as `LCIA Implementation 3.9.xlsx' usualy found in `ecoinvent 3.9_LCIA_implementation.7z'
        biosphere_name: biosphere database name to use as biosphere usualy biosphere3
        version: ecoinvent version string such as "3.9", "3.9.1"; used to apply correct fix/strategy to load the data
    """
    from ecoinvent_interface.string_distance import damerau_levenshtein
    import bw2data
    from bw2io.extractors import ExcelExtractor
    from bw2io.ecoinvent import (
        get_excel_sheet_names,
        header_dict,
        pick_a_unit_label_already,
        drop_unspecified
    )

    if biosphere_name is None:
        biosphere_name = bw2data.config.biosphere
    if biosphere_name not in bw2data.databases or not len(bw2data.Database(biosphere_name)):
        raise ValueError(f"Can't find populated biosphere flow database {biosphere_name}")

    #lcia_file = ei.get_excel_lcia_file_for_version(release=release, version=version)
    lcia_file = path

    sheet_names = get_excel_sheet_names(lcia_file)

    if "units" in sheet_names:
        units_sheetname = "units"
    elif "Indicators" in sheet_names:
        units_sheetname = "Indicators"
    else:
        raise ValueError(
            f"Can't find worksheet for impact category units in {sheet_names}"
        )

    if "CFs" not in sheet_names:
        raise ValueError(
            f"Can't find worksheet for characterization factors; expected `CFs`, found {sheet_names}"
        )

    data = dict(ExcelExtractor.extract(lcia_file))
    units = header_dict(data[units_sheetname])

    cfs = header_dict(data["CFs"])

    CF_COLUMN_LABELS = {
        "3.4": "cf 3.4",
        "3.5": "cf 3.5",
        "3.6": "cf 3.6",
    }
    cf_col_label = CF_COLUMN_LABELS.get(version, "cf")
    units_col_label = pick_a_unit_label_already(units[0])
    units_mapping = {
        (row["method"], row["category"], row["indicator"]): row[units_col_label]
        for row in units
    }

    biosphere_mapping = {}
    for flow in bw2data.Database(biosphere_name):
        biosphere_mapping[(flow["name"],) + tuple(flow["categories"])] = flow.key
        if flow["name"].startswith("[Deleted]"):
            biosphere_mapping[
                (flow["name"].replace("[Deleted]", ""),) + tuple(flow["categories"])
            ] = flow.key

    lcia_data_as_dict = defaultdict(list)

    unmatched = set()
    substituted = set()

    for row in cfs:
        impact_category = (row["method"], row["category"], row["indicator"])
        if row[cf_col_label] is None:
            continue
        try:
            lcia_data_as_dict[impact_category].append(
                (
                    biosphere_mapping[
                        drop_unspecified(
                            row["name"], row["compartment"], row["subcompartment"]
                        )
                    ],
                    float(row[cf_col_label]),
                )
            )
        except KeyError:
            # How is this possible? We are matching ecoinvent data against
            # ecoinvent data from the same release! And yet it moves...
            category = (
                (row["compartment"], row["subcompartment"])
                if row["subcompartment"].lower() != "unspecified"
                else (row["compartment"],)
            )
            same_context = {
                k[0]: v for k, v in biosphere_mapping.items() if k[1:] == category
            }
            candidates = sorted(
                [
                    (damerau_levenshtein(name, row["name"]), name)
                    for name in same_context
                ]
            )
            if (
                candidates[0][0] < 3
                and candidates[0][0] != candidates[1][0]
                and candidates[0][1][0].lower() == row["name"][0].lower()
            ):
                new_name = candidates[0][1]
                pair = (new_name, row["name"])
                if pair not in substituted:
                    print(f"Substituting {new_name} for {row['name']}")
                    substituted.add(pair)
                lcia_data_as_dict[impact_category].append(
                    (
                        same_context[new_name],
                        float(row[cf_col_label]),
                    )
                )
            else:
                if row["name"] not in unmatched:
                    print(
                        "Skipping unmatched flow {}:({}, {})".format(
                            row["name"], row["compartment"], row["subcompartment"]
                        )
                    )
                    unmatched.add(row["name"])

    for key in lcia_data_as_dict:
        method = bw2data.Method(key)
        method.register(
            unit=units_mapping.get(key, "Unknown"),
            filepath=str(lcia_file),
            ecoinvent_version=version,
            database=biosphere_name,
        )
        method.write(lcia_data_as_dict[key])
    pass

def load_ecoinvent_database(path, db_name, biosphere_name):
    import os
    import brightway2 as bw
    import bw2data
    from bw2io.importers import Ecospold2BiosphereImporter, SingleOutputEcospold2Importer
    if biosphere_name not in bw.databases:
        eb = Ecospold2BiosphereImporter(name=biosphere_name,
             filepath=os.path.join(path, "MasterData", "ElementaryExchanges.xml"))
        eb.apply_strategies()
        if not eb.all_linked:
                raise ValueError(f"Can't ingest biosphere database {biosphere_name} - unlinked flows.")
        eb.write_database(overwrite=False)
    bw2data.preferences["biosphere_database"] = biosphere_name

    if db_name not in bw.databases:
        soup = SingleOutputEcospold2Importer(
            dirpath=os.path.join(path, "datasets"),
            db_name=db_name,
            biosphere_database_name=biosphere_name,
            signal=None)
        soup.apply_strategies()
        if not soup.all_linked:
            raise ValueError(f"Can't ingest inventory database {db_name} - unlinked flows.")
        soup.write_database()





# coding=utf-8
#
# Copyright 2021-2022 Romain BESSEAU <romain.besseau@ec.europa.eu>
# Copyright 2022-2024 Alejandra CUE GONZALEZ <alejandra.cue_gonzalez@minesparis.psl.eu>
# Copyright 2022-2024 Benoît GSCHWIND <benoit.gschwind@minesparis.psl.eu>
# Copyright 2022-2024 MINES Paris
#
# This file is part of "parasol-lca" and you can used under the
# term of European Union Public Licence.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# European Union Public Licence for more details.
#
# Version: 22 April 2025
#
# Code authors: Romain BESSEAU, Benoît GSCHWIND, Alejandra CUE GONZALEZ
#

import lca_algebraic as agb

# Electricity dataset

# Switch parameter for electricity mix
Manufacturing_electricity_mix = agb.newEnumParam(
    name = 'Manufacturing_electricity_mix',
    values = ['FR', 'EU', 'US', 'CN','World', 'PV', 'Nuclear','Coal', 'CO2_content'],
    default = 'World',
    description = "Switch on electricty mix used for PV manufacture",
    group = 'PV manufacture')

Electricity_mix_CO2_content = agb.newFloatParam(
    name = "Electricity_mix_CO2_content",
    default = 0.5, min = 0, max = 1,
    distrib = agb.params.DistributionType.TRIANGLE,
    group = "PV manufacture",
    unit = "ratio or quasi-equivalent in kgCO2eq/kWh")

# Mounting System

# Define parameters for the mounting system
roof_ratio = agb.newFloatParam(
    name = "roof_vs_ground_ratio",
    default = 1.0, min = 0.0, max = 1.0,
    distrib = agb.params.DistributionType.LINEAR,
    group = "PV install",
    description = "Proportion of rooftop installations",
    unit = 'fraction')

mounting_system_weight_total = agb.newFloatParam(
    name = "Mounting_system_weight_total",
    default = 5, min = 2, max = 11.5,
    distrib = agb.params.DistributionType.TRIANGLE,
    group = "PV intall",
    unit = "kg/m²")

mounting_system_weight_alu = agb.newFloatParam(
    name = "Mounting_system_weight_alu",
    default = 1.5, min = 0.67, max = 2.4,
    distrib = agb.params.DistributionType.TRIANGLE,
    group = "PV install",
    unit = "kg/m²")

mounting_system_weight_wood = agb.newFloatParam(
    name = "Mounting_system_weight_wood",
    default = 0, min = 0, max = 10,
    distrib = agb.params.DistributionType.TRIANGLE,
    group = "PV install",
    unit = "kg/m²")


ground_coverage_ratio = agb.newFloatParam(
    name = 'Ground_coverage_ratio',
    default = 0.4, min = 0.2, max = 0.5,
    distrib = agb.params.DistributionType.TRIANGLE,
    group = "PV install",
    unit = "fraction",
    description = "Ground coverage ratio")

lifetime = agb.newFloatParam(
    name = 'Power_plant_lifetime',
    default = 30, min = 20, max = 40,
    distrib = agb.params.DistributionType.TRIANGLE,
    group = "PV operation",
    unit = "year",
    description = "Lifetime of the PV plant")

# Electrical installation

P_install = agb.newFloatParam(
    name = "Power_plant_capacity",
    default = 100,
    min = 3, max = 10000,
    group = "PV install", unit = "kWp")

recycling_rate = agb.newFloatParam(
    name = "Recycling_rate",
    default=0.9, min=0, max=1,
    group = "PV recycle",
    unit="fraction")

# Silicon production

#parameter to adjust the amount of electricity for silicon manufacturing
silicon_elec_intensity = agb.newFloatParam(
    name = "Silicon_production_electricity_intensity",
    default = 30, min = 11, max = 180,
    distrib = agb.DistributionType.TRIANGLE,
    group = "PV manufacture",
    unit = "kWh/kg")

silicon_heat_intensity = agb.newFloatParam(
    name = "Silicon_production_heat_intensity",
    default = 185, min = 0, max = 185,
    distrib = agb.DistributionType.TRIANGLE,
    group = "PV manufacture",
    unit = "MJ/kg")

silicon_casting_elec_intensity = agb.newFloatParam(
    name = "Silicon_casting_electricity_intensity",
    default = 15, min = 10, max = 30,
    distrib = agb.DistributionType.TRIANGLE,
    group = "PV manufacture",
    unit = "kWh/kg")

## Wafer

# Parameters
diamond_wiring = agb.newBoolParam(
    name = "Diamond_wiring_cutting",
    default = 1,
    distrib = agb.DistributionType.LINEAR,
    group = "PV manufacture")

wafer_thickness = agb.newFloatParam(
    name = "Wafer_thickness",
    default=160, min=128, max=190,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV manufacture",
    unit="µm")

kerf_loss = agb.newFloatParam(
    name = "Kerf_loss",
    default = 0.3, min=0.3, max=0.50,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV manufacture",
    unit="fraction")

sic_recycled_share = agb.newFloatParam(
    name = "SiC_recycled_share",
    default = 0.0, min=0.0, max=0.9,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV install",
    unit="fraction")

manufacturing_efficiency = agb.newFloatParam(
    name = "Manufacturing_efficiency_gains",
    default = 0, min=0, max=1,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV manufacture",
    description="Manufacturing efficiency gain (elec for wafer)",
    unit="fraction")


# PV Cell manufacturing

silver_amount = agb.newFloatParam(
    name = "Silver_content",
    default = 9.6, min=2, max=9.6,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV manufacture",
    description="Silver amount for PV cell", unit="g/m²")

# PV Panel

m_aluminium_frame = agb.newFloatParam(
    name = "Aluminium_frame_surfacic_weight",
    default = 1.5, min=0, max=2.63,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV install",
    unit="kg/m2")

bifaciale = agb.newBoolParam(
    name = "Bifaciale_modules",
    default = False,
    group="PV install")

Recycling_rate_Al = agb.newFloatParam(
    name = "Recycling_rate_Al",
    default = 0.96, min=0.56, max=1,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV recycle",
    unit="fraction")

Recycling_rate_Cu = agb.newFloatParam(
    name = "Recycling_rate_Cu",
    default = 0.75, min=0.44, max=0.96,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV recycle",
    unit="fraction")

recycling_rate_glass = agb.newFloatParam(
    name = "Recycling_rate_glass",
    default = 0.9, min=0.6, max=1,
    distrib = agb.DistributionType.TRIANGLE,
    group = "PV recycle",
    unit="fraction")

electricity_recycling = agb.newFloatParam(
    name = "Electricity_consumption_for_recycling",
    default = 50, min=0, max=250,
    distrib = agb.DistributionType.TRIANGLE,
    group = "PV recycle",
    description="elec used for recycling", unit="kWh/t")

heat_recycling = agb.newFloatParam(
    name = "Heat_consumption_for_recycling",
    default = 76, min=0, max=150,
    distrib = agb.DistributionType.TRIANGLE,
    group = "PV recycle",
    description="Heat for recycling", unit="MJ/t")

glass_thickness = agb.newFloatParam(
    name = "Glass_thickness",
    default = 4, min=2, max=4,
    distrib = agb.DistributionType.TRIANGLE,
    group="PV install",
    unit = "mm")

# PV System

module_efficiency =  agb.newFloatParam(
    name = "PV_module_efficiency",
    default=0.22, min=0.15, max=0.23,
    distrib = agb.DistributionType.TRIANGLE,
    group = "PV install",
    unit="kWp/m²")

inverter_lifetime = agb.newFloatParam(
    name = "Inverter_lifetime",
    group = "PV operation",
    default=15, min=10, max=30,
    distrib = agb.DistributionType.TRIANGLE,
    unit="year")

inverter_weight_per_kW = agb.newFloatParam(
    name = "Inverter_weight_per_kW",
    default=2, min=1, max=6,
    group = 'PV install',
    label_fr="poids onduleur",
    unit="kg/kWp")

d_lorry = agb.newFloatParam(
    name = "Transport_distance_lorry",
    default = 1000, min=40, max=2000,
    unit="km",
    group="PV transport")

d_train = agb.newFloatParam(
    name = "Transport_distance_train",
    default=500, min=100, max=600,
    unit="km",
    group="PV transport")

d_sea = agb.newFloatParam(
    name = "Transport_distance_boat",
    default=5000, min=2000, max=6000,
    unit="km",
    group="PV transport")

electrical_installation_weight_per_kW = agb.newFloatParam(
    name = "Electrical_installation_specific_weight",
    default = 3, min=2.15, max=4.6,
    unit="kg/kW",
    group="PV install")

kWhperkWp = agb.newFloatParam(
    name = 'Normalised_annual_PV_production_kWh_per_kWp',
    default = 1300, min = 900, max = 1700,
    distrib = agb.params.DistributionType.TRIANGLE,
    group = "PV operation",
    unit = "kWh/kWp/year",
    description = "Durée de vie")

